import numpy as np

class Tensor:
    def __init__(self, data, parents=(), operation=None):
        data = np.array(data, dtype=np.float64) if not isinstance(data, np.ndarray) else data
        self.data = data
        self.parents = set(parents)
        self.grad = np.zeros_like(data)
        self.operation = operation
        self._backward = lambda: None
    
    def __add__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data + other.data, parents=(self, other), operation='+')
        def _backward():
            if len(self.parents) != 0:
                self.grad += 1.0 * out.grad
            if len(other.parents) != 0:    
                other.grad += 1.0 * out.grad
        out._backward = _backward
        return out
    
    def __radd__(self, other):
        return self + other

    def __mul__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data * other.data, parents=(self, other), operation='*')
        def _backward():
            if len(self.parents) != 0:
                self.grad += other.data * out.grad
            if len(other.parents) != 0:
                other.grad += self.data * out.grad
        out._backward = _backward
        return out
    
    def __matmul__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data @ other.data, parents=(self, other), operation='@')
        def _backward():
            self.grad += out.grad @ other.data.T
            other.grad += self.data.T @ out.grad
        out._backward = _backward
        return out
    
    def __rmul__(self, other):
        return self * other
    
    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return Tensor(other) - self

    def __truediv__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        out = Tensor(self.data / other.data, parents=(self, other), operation='/')
        def _backward():
            if len(self.parents) != 0:
                self.grad += (1.0 / other.data) * out.grad

            if len(other.parents) != 0:
                other.grad += (-self.data / (other.data ** 2)) * out.grad

        out._backward = _backward
        return out

    def __rtruediv__(self, other):
        return Tensor(other) / self
    
    def __pow__(self, other):
        assert isinstance(other, (int, float)), "only supporting int/float powers for now"
        out = Tensor(self.data**other, (self,), operation='**')
        def _backward():
            self.grad += (other * self.data**(other-1)) * out.grad
        out._backward = _backward
        return out
    
    def relu(self):
        out = Tensor(np.maximum(0, self.data), parents=(self,), operation='relu')
        def _backward():
            self.grad += (self.data > 0).astype(self.data.dtype) * out.grad
        out._backward = _backward
        return out
    
    def sigmoid(self):
        s = 1 / (1 + np.exp(-self.data))
        out = Tensor(s, parents=(self,), operation='sigmoid')
        def _backward():
            self.grad += s * (1 - s) * out.grad
        out._backward = _backward
        return out
    
    def tanh(self):
        t = np.tanh(self.data)
        out = Tensor(t, parents=(self,), operation='tanh')
        def _backward():
            self.grad += (1 - t ** 2) * out.grad
        out._backward = _backward
        return out
    
    def __neg__(self):
        return self * -1
    
    def backward(self):
        topology = []
        visited = set()
        def build_topology(tensor):
            if tensor not in visited:
                visited.add(tensor)
                for parent in tensor.parents:
                    build_topology(parent)
                topology.append(tensor)
        build_topology(self)
        
        self.grad = np.ones_like(self.data)
        for tensor in reversed(topology):
            tensor._backward()

    def mean(self):
        n = self.data.size
        out = Tensor(np.array(self.data.mean()), parents=(self,), operation='mean')
        def _backward():
            self.grad += (1.0 / n) * np.ones_like(self.data) * out.grad
        out._backward = _backward
        return out


from ..tensor import Tensor
import numpy as np

class Linear:
    def __init__(self, size, input_size=None):
        self.n = size
        self.input_size = input_size
        self.weights = None
        self.bias = None

    def __call__(self, x):
        return self.weights @ x + self.bias
    
    def compile(self):
        self.weights = Tensor(np.random.uniform(-1, 1, size=(self.n, self.input_size)))
        self.bias = Tensor(np.random.uniform(-1, 1, size=(self.n, 1)))

class Activation:
    def __init__(self):
        self.n = None

class Tanh(Activation):
    def __init__(self):
        super().__init__()

    def __call__(self, x):
        return x.tanh()
    
class Relu(Activation):
    def __init__(self):
        super().__init__()

    def __call__(self, x):
        return x.relu()



from ..layers.layers import Activation
import numpy as np

class SGD:
    def __init__(self, alpha=0.01, model=None, auto_zero_grad=True):
        self.model = model
        self.alpha = alpha
        self.auto_zero_grad = auto_zero_grad

    def step(self):
        assert self.model_is_set(), "Model must be set before calling the step function."
        for layer in self.model.layers:
            if isinstance(layer, Activation):
                continue
            layer.weights.data -= self.alpha * layer.weights.grad
            layer.bias.data -= self.alpha * layer.bias.grad
            if self.auto_zero_grad:
                layer.weights.grad = np.zeros_like(layer.weights.data)
                layer.bias.grad = np.zeros_like(layer.bias.data)

    def zero_grad(self):
        assert self.model_is_set(), "Model must be set before calling the zero_grad function."
        for layer in self.model.layers:
            layer.weights.grad = np.zeros_like(layer.weights.data)
            layer.bias.grad = np.zeros_like(layer.bias.data)
            
    def model_is_set(self):
        return self.model is not None



from ..layers.layers import Activation
import numpy as np

class Sequential:
    def __init__(self, *layers):
        self.layers = layers
        self.compile_layers(layers)

    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x
    
    def compile_layers(self, layers):
        for i, layer in enumerate(layers):
            assert self.first_layer_has_input_size(i), f"Layer 0 must have input_size defined."
            if self.is_activation_layer(layer):
                layer.n = layers[i-1].n
                continue
            layer.input_size = layer.input_size if self.is_first_layer(i) else layers[i-1].n
            layer.compile()

    def is_first_layer(self, index):
        return index == 0
    
    def is_activation_layer(self, layer):
        return isinstance(layer, Activation)
    
    def first_layer_has_input_size(self, index):
        if self.is_first_layer(index):
            return self.layers[index].input_size is not None
        return True
    
    def predict(self, x):
        x = x.T
        with np.errstate(divide='ignore', invalid='ignore'):
            for layer in self.layers:
                x = layer(x)
        return x.data
    
    def load_data(self, batch_size, X, Y, shuffle=True):
        num_examples = X.shape[0]
        
        if shuffle:
            idx = np.random.permutation(num_examples)
            X = X[idx]
            Y = Y[idx]

        for i in range(0, num_examples, batch_size):
            X_batch = X[i:i+batch_size]
            Y_batch = Y[i:i+batch_size]
            yield X_batch, Y_batch
    
    def train(self, X, Y, optimizer, loss_fn, epochs=100, batch_size=128, print_every=50):
        optimizer.model = self
        for epoch in range(1, epochs+1):
            for X_batch, Y_batch in self.load_data(batch_size, X, Y):
                logits = self(X_batch.T)
                loss = loss_fn(logits, Y_batch.T)

                loss.backward()
                optimizer.step()
            
            if epoch % print_every == 0:
                print(f"(Epoch {epoch}) Loss: {loss.data}")
